package org.commonjava.depbase.web.base.rest;

import static javax.ws.rs.core.Response.Status.BAD_REQUEST;
import static javax.ws.rs.core.Response.Status.NO_CONTENT;
import static org.apache.maven.graph.common.RelationshipType.DEPENDENCY;
import static org.apache.maven.graph.common.RelationshipType.EXTENSION;
import static org.apache.maven.graph.common.RelationshipType.PLUGIN;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.enterprise.context.RequestScoped;
import javax.inject.Inject;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;

import org.apache.maven.graph.common.DependencyScope;
import org.apache.maven.graph.common.RelationshipType;
import org.apache.maven.graph.common.ref.ProjectVersionRef;
import org.apache.maven.graph.common.version.InvalidVersionSpecificationException;
import org.apache.maven.graph.effective.rel.DependencyRelationship;
import org.apache.maven.graph.effective.rel.ProjectRelationship;
import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.data.TensorDataManager;
import org.commonjava.tensor.inject.TensorData;
import org.commonjava.tensor.util.ProjectVersionRefComparator;
import org.commonjava.util.logging.Logger;
import org.commonjava.web.json.model.Listing;
import org.commonjava.web.json.ser.JsonSerializer;

@Path( "/rel/direct" )
@Produces( MediaType.APPLICATION_JSON )
@RequestScoped
public class DirectRelationshipResource
{
    private final Logger logger = new Logger( getClass() );

    @Inject
    private TensorDataManager data;

    @Inject
    @TensorData
    private JsonSerializer serializer;

    @Path( "/list" )
    @GET
    public Response list( @QueryParam( "g" ) final String groupIdPattern,
                          @QueryParam( "a" ) final String artifactIdPattern )
    {
        Response response = Response.status( NO_CONTENT )
                                    .build();

        try
        {
            final Set<ProjectVersionRef> refs = data.getAllStoredProjectRefs();
            final List<ProjectVersionRef> list = new ArrayList<ProjectVersionRef>();
            if ( refs != null )
            {
                if ( groupIdPattern != null || artifactIdPattern != null )
                {
                    final String gip = groupIdPattern == null ? ".*" : groupIdPattern.replaceAll( "\\*", ".*" );
                    final String aip = artifactIdPattern == null ? ".*" : artifactIdPattern.replaceAll( "\\*", ".*" );

                    logger.info( "Filtering %d projects using groupId pattern: '%s' and artifactId pattern: '%s'",
                                 refs.size(), gip, aip );

                    for ( final ProjectVersionRef ref : refs )
                    {
                        if ( ref.getGroupId()
                                .matches( gip ) && ref.getArtifactId()
                                                      .matches( aip ) )
                        {
                            list.add( ref );
                        }
                    }
                }
                else
                {
                    logger.info( "Returning all %d projects", refs.size() );
                    list.addAll( refs );
                }

            }

            if ( !list.isEmpty() )
            {
                Collections.sort( list, new ProjectVersionRefComparator() );

                final String json = serializer.toString( new Listing<ProjectVersionRef>( list ) );
                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            response = Response.serverError()
                               .build();
        }

        return response;
    }

    @Path( "/{g}/{a}/{v}/parent" )
    @GET
    public Response parentOf( @PathParam( "g" ) final String groupId, @PathParam( "a" ) final String artifactId,
                              @PathParam( "v" ) final String version )
    {
        Response response = Response.status( NO_CONTENT )
                                    .build();

        // TODO: Discover graph itself if missing, if discovery allowed...
        try
        {
            final ProjectVersionRef parent = data.getParent( new ProjectVersionRef( groupId, artifactId, version ) );
            if ( parent != null )
            {
                final String json = serializer.toString( parent );
                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup parent for: %s:%s:%s. Reason: %s", e, groupId, artifactId, version,
                          e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );

            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

    @Path( "/{g}/{a}/{v}/dependencies" )
    @GET
    public Response dependenciesOf( @PathParam( "g" ) final String groupId, @PathParam( "a" ) final String artifactId,
                                    @PathParam( "v" ) final String version,
                                    @QueryParam( "scopes" ) final String scopesStr )
    {
        final Set<DependencyScope> acceptedScopes = new HashSet<DependencyScope>();
        if ( scopesStr != null && scopesStr.trim()
                                           .length() > 0 )
        {
            final String[] scopes = scopesStr.split( "\\s*,\\s*" );
            for ( final String scope : scopes )
            {
                final DependencyScope s = DependencyScope.getScope( scope );
                if ( s != null )
                {
                    acceptedScopes.add( s );
                }
            }
        }

        if ( acceptedScopes.isEmpty() )
        {
            acceptedScopes.addAll( Arrays.asList( DependencyScope.values() ) );
        }

        Response response = Response.status( NO_CONTENT )
                                    .build();

        // TODO: Filtering using similar setup to the graph resource...
        // TODO: Discover graph itself if missing, if discovery allowed...
        try
        {
            final Set<ProjectRelationship<?>> deps =
                data.getAllDirectRelationshipsWithExactSource( new ProjectVersionRef( groupId, artifactId, version ),
                                                               DEPENDENCY );

            if ( deps != null )
            {
                for ( final Iterator<ProjectRelationship<?>> it = deps.iterator(); it.hasNext(); )
                {
                    final DependencyRelationship rel = (DependencyRelationship) it.next();
                    if ( !acceptedScopes.contains( rel.getScope() ) )
                    {
                        it.remove();
                    }
                }

                final String json = serializer.toString( new Listing<ProjectRelationship<?>>( deps ) );
                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup dependencies for: %s:%s:%s. Reason: %s", e, groupId, artifactId, version,
                          e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );

            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

    @Path( "/{g}/{a}/{v}/plugins" )
    @GET
    public Response pluginsOf( @PathParam( "g" ) final String groupId, @PathParam( "a" ) final String artifactId,
                               @PathParam( "v" ) final String version )
    {
        Response response = Response.status( NO_CONTENT )
                                    .build();

        // TODO: Filtering using similar setup to the graph resource...
        // TODO: Discover graph itself if missing, if discovery allowed...
        try
        {
            final Set<ProjectRelationship<?>> deps =
                data.getAllDirectRelationshipsWithExactSource( new ProjectVersionRef( groupId, artifactId, version ),
                                                               PLUGIN );

            if ( deps != null )
            {
                final String json = serializer.toString( new Listing<ProjectRelationship<?>>( deps ) );
                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup plugins for: %s:%s:%s. Reason: %s", e, groupId, artifactId, version,
                          e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );

            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

    @Path( "/{g}/{a}/{v}/extensions" )
    @GET
    public Response extensionsOf( @PathParam( "g" ) final String groupId, @PathParam( "a" ) final String artifactId,
                                  @PathParam( "v" ) final String version )
    {
        Response response = Response.status( NO_CONTENT )
                                    .build();

        // TODO: Filtering using similar setup to the graph resource...
        // TODO: Discover graph itself if missing, if discovery allowed...
        try
        {
            final Set<ProjectRelationship<?>> deps =
                data.getAllDirectRelationshipsWithExactSource( new ProjectVersionRef( groupId, artifactId, version ),
                                                               EXTENSION );

            if ( deps != null )
            {
                final String json = serializer.toString( new Listing<ProjectRelationship<?>>( deps ) );
                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup extensions for: %s:%s:%s. Reason: %s", e, groupId, artifactId, version,
                          e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );

            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

    @Path( "/{g}/{a}/{v}/relationships" )
    @GET
    public Response relationshipsSpecifiedBy( @PathParam( "g" ) final String groupId,
                                              @PathParam( "a" ) final String artifactId,
                                              @PathParam( "v" ) final String version,
                                              @QueryParam( "referenceTypes" ) final String refTypes )
    {
        final Set<RelationshipType> acceptedTypes = new HashSet<RelationshipType>();
        if ( refTypes != null )
        {
            if ( refTypes != null && refTypes.trim()
                                             .length() > 0 )
            {
                final String[] types = refTypes.split( "\\s*,\\s*" );
                for ( final String type : types )
                {
                    final RelationshipType t = RelationshipType.getType( type );
                    if ( t != null )
                    {
                        acceptedTypes.add( t );
                    }
                }
            }
        }

        Response response = Response.status( NO_CONTENT )
                                    .build();

        // TODO: Filtering using similar setup to the graph resource...
        // TODO: Discover graph itself if missing, if discovery allowed...
        try
        {
            final Set<ProjectRelationship<?>> rels =
                data.getAllDirectRelationshipsWithExactSource( new ProjectVersionRef( groupId, artifactId, version ),
                                                               acceptedTypes.toArray( new RelationshipType[] {} ) );

            if ( rels != null )
            {
                final String json = serializer.toString( new Listing<ProjectRelationship<?>>( rels ) );
                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup relationships specified by: %s:%s:%s. Reason: %s", e, groupId, artifactId,
                          version, e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );

            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

    @Path( "/{g}/{a}/{v}/users" )
    @GET
    public Response relationshipsTargeting( @PathParam( "g" ) final String groupId,
                                            @PathParam( "a" ) final String artifactId,
                                            @PathParam( "v" ) final String version,
                                            @QueryParam( "referenceTypes" ) final String refTypes )
    {
        final Set<RelationshipType> acceptedTypes = new HashSet<RelationshipType>();
        if ( refTypes != null )
        {
            if ( refTypes != null && refTypes.trim()
                                             .length() > 0 )
            {
                final String[] types = refTypes.split( "\\s*,\\s*" );
                for ( final String type : types )
                {
                    final RelationshipType t = RelationshipType.getType( type );
                    if ( t != null )
                    {
                        acceptedTypes.add( t );
                    }
                }
            }
        }

        Response response = Response.status( NO_CONTENT )
                                    .build();

        // TODO: Filtering using similar setup to the graph resource...
        // TODO: Discover graph itself if missing, if discovery allowed...
        try
        {
            final Set<ProjectRelationship<?>> rels =
                data.getAllDirectRelationshipsWithExactTarget( new ProjectVersionRef( groupId, artifactId, version ),
                                                               acceptedTypes.toArray( new RelationshipType[] {} ) );

            if ( rels != null )
            {
                final String json = serializer.toString( new Listing<ProjectRelationship<?>>( rels ) );
                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup relationships specified by: %s:%s:%s. Reason: %s", e, groupId, artifactId,
                          version, e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );

            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

}
