package org.commonjava.depbase.web.base.rest;

import static javax.ws.rs.core.Response.Status.BAD_REQUEST;
import static javax.ws.rs.core.Response.Status.NO_CONTENT;
import static org.commonjava.depbase.web.base.rest.RequestUtils.createAggregationOptions;

import java.util.List;

import javax.enterprise.context.RequestScoped;
import javax.inject.Inject;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;

import org.apache.maven.graph.common.DependencyScope;
import org.apache.maven.graph.common.ref.ProjectVersionRef;
import org.apache.maven.graph.common.version.InvalidVersionSpecificationException;
import org.apache.maven.graph.effective.EProjectGraph;
import org.apache.maven.graph.effective.transform.TransitiveDependencyTransformer;
import org.apache.maven.graph.effective.traverse.BuildOrderTraversal;
import org.apache.maven.graph.effective.traverse.PrintingTraversal;
import org.apache.maven.graph.effective.traverse.model.BuildOrder;
import org.commonjava.tensor.agg.AggregationOptions;
import org.commonjava.tensor.agg.GraphAggregator;
import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.data.TensorDataManager;
import org.commonjava.tensor.discover.ProjectRelationshipDiscoverer;
import org.commonjava.tensor.inject.TensorData;
import org.commonjava.util.logging.Logger;
import org.commonjava.web.json.model.Listing;
import org.commonjava.web.json.ser.JsonSerializer;

@Path( "/rel/graph" )
@Produces( MediaType.APPLICATION_JSON )
@RequestScoped
public class GraphRelationshipResource
{
    private final Logger logger = new Logger( getClass() );

    @Inject
    private TensorDataManager data;

    @Inject
    private GraphAggregator aggregator;

    @Inject
    private ProjectRelationshipDiscoverer relationshipDiscoverer;

    @Inject
    @TensorData
    private JsonSerializer serializer;

    @Path( "/ancestry/{g}/{a}/{v}" )
    @GET
    public Response ancestryOf( @PathParam( "g" ) final String groupId, @PathParam( "a" ) final String artifactId,
                                @PathParam( "v" ) final String version )
    {
        Response response = Response.status( NO_CONTENT )
                                    .build();
        try
        {
            final List<ProjectVersionRef> ancestry =
                data.getAncestry( new ProjectVersionRef( groupId, artifactId, version ) );

            if ( ancestry != null )
            {
                final String json = serializer.toString( new Listing<ProjectVersionRef>( ancestry ) );
                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup ancestry for: %s:%s:%s. Reason: %s", e, groupId, artifactId, version,
                          e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );
            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

    @Path( "/build-order/{g}/{a}/{v}" )
    @GET
    public Response buildOrder( @PathParam( "g" ) final String groupId, @PathParam( "a" ) final String artifactId,
                                @PathParam( "v" ) final String version, @Context final HttpServletRequest request )
    {
        Response response = Response.status( NO_CONTENT )
                                    .build();
        try
        {
            final AggregationOptions options = createAggregationOptions( request );

            final ProjectVersionRef ref = new ProjectVersionRef( groupId, artifactId, version );
            EProjectGraph graph = data.getProjectGraph( ref );
            if ( graph == null && options.isDiscoveryEnabled() )
            {
                relationshipDiscoverer.discoverRelationships( ref, options.getDiscoveryConfig() );
                graph = data.getProjectGraph( ref );
            }

            if ( graph != null )
            {
                graph = aggregator.connectSubgraphs( graph, options, true );

                final BuildOrderTraversal traversal = new BuildOrderTraversal( options.getFilter() );

                graph.traverse( traversal );

                final BuildOrder buildOrder = traversal.getBuildOrder();
                final String json = serializer.toString( buildOrder );

                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup project graph for: %s:%s:%s. Reason: %s", e, groupId, artifactId, version,
                          e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );
            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

    @Path( "/project/{g}/{a}/{v}" )
    @GET
    public Response projectGraph( @PathParam( "g" ) final String groupId, @PathParam( "a" ) final String artifactId,
                                  @PathParam( "v" ) final String version, @Context final HttpServletRequest request )
    {
        Response response = Response.status( NO_CONTENT )
                                    .build();
        try
        {
            final AggregationOptions options = createAggregationOptions( request );

            final ProjectVersionRef ref = new ProjectVersionRef( groupId, artifactId, version );
            EProjectGraph graph = data.getProjectGraph( ref );
            if ( graph == null && options.isDiscoveryEnabled() )
            {
                relationshipDiscoverer.discoverRelationships( ref, options.getDiscoveryConfig() );
                graph = data.getProjectGraph( ref );
            }

            if ( graph != null )
            {
                graph = aggregator.connectSubgraphs( graph, options, true );

                final String json = serializer.toString( graph );
                response = Response.ok( json )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup project graph for: %s:%s:%s. Reason: %s", e, groupId, artifactId, version,
                          e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );
            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

    @Path( "/tree/{g}/{a}/{v}" )
    @Produces( MediaType.TEXT_PLAIN )
    @GET
    public Response depTree( @PathParam( "g" ) final String groupId, @PathParam( "a" ) final String artifactId,
                             @PathParam( "v" ) final String version, @Context final HttpServletRequest request,
                             @QueryParam( "s" ) @DefaultValue( "runtime" ) final String scope,
                             @QueryParam( "c-t" ) @DefaultValue( "true" ) final boolean collapseTransitives )
    {
        Response response = Response.status( NO_CONTENT )
                                    .build();
        try
        {
            final AggregationOptions options = createAggregationOptions( request );

            final ProjectVersionRef ref = new ProjectVersionRef( groupId, artifactId, version );
            EProjectGraph graph = data.getProjectGraph( ref );
            if ( graph == null && options.isDiscoveryEnabled() )
            {
                relationshipDiscoverer.discoverRelationships( ref, options.getDiscoveryConfig() );
                graph = data.getProjectGraph( ref );
            }

            if ( graph != null )
            {
                graph = aggregator.connectSubgraphs( graph, options, true );

                if ( collapseTransitives )
                {
                    final TransitiveDependencyTransformer trans =
                        new TransitiveDependencyTransformer( DependencyScope.getScope( scope ) );
                    graph.traverse( trans );

                    graph = trans.getTransformedGraph();
                }

                final PrintingTraversal printer = new PrintingTraversal();
                graph.traverse( printer );

                response = Response.ok( printer.getString() )
                                   .build();
            }
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to lookup project graph for: %s:%s:%s. Reason: %s", e, groupId, artifactId, version,
                          e.getMessage() );

            response = Response.serverError()
                               .build();
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Invalid version in request: '%s'. Reason: %s", e, version, e.getMessage() );
            response = Response.status( BAD_REQUEST )
                               .entity( "Invalid version: '" + version + "'" )
                               .build();
        }

        return response;
    }

}
